import os
import sys
import random
from abc import ABC, abstractmethod

import torch
from torch import nn
import numpy as np
from mpi4py import MPI
import gym
from pipcs import Config

from .utils import *


class Policy(nn.Module, ABC):
    """Abstract subclass of nn.Module."""
    def __init__(self):
        super().__init__()

    @abstractmethod
    def rollout(self, env):
        """This function should be implemented by the user and
        it should evaluate the model in the given environment and
        return the total reward."""
        pass


class NES():
    """
    :ivar int gen: Current generation
    :ivar Policy policy: Trained policy
    :ivar torch.optim.Optimizer optim: Optimizer of the policy
    """
    @hook
    def __init__(self, config: Config):
        config.check_config()
        self.config = config

        if config.nes.seed is not None:
            torch.manual_seed(config.nes.seed)
            np.random.seed(config.nes.seed)
            random.seed(config.nes.seed)
            torch.cuda.manual_seed(config.nes.seed)
            torch.cuda.manual_seed_all(config.nes.seed)
            torch.backends.cudnn.benchmark = False
            torch.backends.cudnn.deterministic = True
            self.env = self.make_env(**config.environment.to_dict())
            self.env.seed(config.nes.seed)
            self.env.action_space.seed(config.nes.seed)

        self.gen = 0

        comm = MPI.COMM_WORLD
        self.n_workers = comm.Get_size()
        self.rank = comm.Get_rank()

        self.policy = self.make_policy(**config.policy.to_dict())
        self.optim = self.make_optimizer(policy=self.policy, **config.optimizer.to_dict())
        self.dummy_policy = self.make_policy(**self.config.policy.to_dict())

    @staticmethod
    def make_env(make_env, **kwargs):
        """Helper function to create a gym environment."""
        return make_env(**kwargs)

    @staticmethod
    def make_policy(policy, **kwargs):
        """Helper function to create a policy."""
        assert issubclass(policy, Policy)
        return policy(**kwargs)

    @staticmethod
    def make_optimizer(policy, optim_type, **kwargs):
        """Helper function to create an optimizer."""
        return optim_type(policy.parameters(), **kwargs)

    def eval_policy(self, policy):
        """Evaluate policy on the ``self.env`` for ``self.config.nes.n_rollout times``"""
        total_reward = 0
        for _ in range(self.config.nes.n_rollout):
            total_reward += policy.rollout(self.env)
        return total_reward / self.config.nes.n_rollout

    @hook
    def sample(self, mean):
        normal = torch.distributions.normal.Normal(0, self.config.nes.sigma)
        epsilon = normal.sample([int(self.config.nes.population_size/2), mean.shape[0]])
        population_params = torch.cat((mean + epsilon, mean - epsilon))
        epsilons = torch.cat((epsilon, -epsilon))
        return population_params, epsilons

    @hook
    def evaluate(self, population_params):
        comm = MPI.COMM_WORLD
        rewards = []
        reward_array = np.zeros(self.config.nes.population_size, dtype=np.float32)
        batch = np.array_split(population_params, self.n_workers)[self.rank]
        for param in batch:
            torch.nn.utils.vector_to_parameters(param, self.dummy_policy.parameters())
            rewards.append(self.eval_policy(self.dummy_policy))
        comm.Allgatherv([np.array(rewards, dtype=np.float32), MPI.FLOAT], [reward_array, MPI.FLOAT])
        return reward_array

    @hook
    def calculate_gradients(self, rewards, mean, epsilons):
        ranked_rewards = torch.from_numpy(rank_transformation(rewards)).unsqueeze(0)
        grad = -(torch.mm(ranked_rewards, epsilons) / (len(rewards) * self.config.nes.sigma))
        grad = (grad + mean * self.config.nes.l2_decay).squeeze()  # L2 Decay
        return grad

    @hook
    def optimize(self, grad):
        self.optim.zero_grad()
        index = 0
        for parameter in self.policy.parameters():
            size = np.prod(parameter.shape)
            parameter.grad = grad[index:index+size].view(parameter.shape)
            # Limit gradient update to increase stability.
            parameter.grad.data.clamp_(-1.0, 1.0)
            index += size
        self.optim.step()

    def train(self):
        """Train ``self.policy`` for ``self.config.nes.n_steps`` to increase reward returns
        from the ``self.env`` using Natural Evolution Strategy gradient estimation."""
        torch.set_grad_enabled(False)
        if not self.rank == 0:
            f = open(os.devnull, 'w')
            sys.stdout = f

        for gen in range(self.config.nes.n_step):
            self.gen = gen

            # Sample
            mean = torch.nn.utils.parameters_to_vector(self.policy.parameters())
            population_params, epsilons = self.sample(mean)

            # Evaluate Population
            rewards = self.evaluate(population_params)

            # Calculate Gradients
            grad = self.calculate_gradients(rewards, mean, epsilons)

            self.optimize(grad)

        sys.stdout = sys.__stdout__
        torch.set_grad_enabled(True)



from typing import Optional, Type, Callable, Any

import torch
from pipcs import Config, Required, required

from .nes import Policy, NES


default_config = Config()


@default_config('environment')
class EnvironmentConfig():
    """**name: environment**

    :ivar Callable[[Any], gym.Env] id: Function that returns a gym environment
    """
    make_env: Required[Callable[[Any], 'gym.Env']] = required


@default_config('policy')
class PolicyConfig():
    """**name: policy**

    :ivar Required[Type[Policy]] policy: torch.nn.Module with a rollout method
    """
    policy: Required[Type[Policy]] = required


@default_config('optimizer')
class OptimizerConfig():
    """**name: optimizer**

    :ivar float lr: Learning rate
    :ivar Required[Type[torch.optim.Optimizer]] optim_type: torch optim module
    """
    lr: float = 0.02
    optim_type: Required[Type[torch.optim.Optimizer]] = required


@default_config('nes')
class NESConfig():
    """**name: nes**

    :ivar int population_size: Population Size, higher means lower variance in
        gradient calculation but higher memory consumption and longer training time.
    :ivar int n_step: Number of training steps
    :ivar float sigma: Standart deviation for population sampling
    :ivar int n_rollout: Number of episodes per sampled policy.
    :ivar Optional[int] seed: Random seed
    """
    n_rollout: int = 1
    n_step: Required[int] = required
    l2_decay: float = 0.005
    population_size: Required[int] = required
    sigma: float = 0.02
    seed: Optional[int] = None



import gym
from pipcs import Config
import torch
from torch import nn

from nes import NES, Policy, default_config


class Agent(Policy):
    def __init__(self):
        super().__init__()
        self.seq = nn.Sequential(
            nn.Linear(4, 64),
            nn.ReLU(),
            nn.Linear(64, 32),
            nn.ReLU(),
            nn.Linear(32, 2))

    def rollout(self, env):
        done = False
        obs = env.reset()
        total_reward = 0
        while not done:
            obs_tensor = torch.from_numpy(obs).float()
            action = self.seq(obs_tensor).argmax().item()
            obs, reward, done, _ = env.step(action)
            total_reward += reward
        return total_reward


config = Config(default_config)


@config('environment')
class EnvironmentConfig():
    make_env = gym.make
    id: str = 'CartPole-v1'


@config('policy')
class PolicyConfig():
    policy = Agent


@config('optimizer')
class OptimizerConfig():
    lr = 0.02
    optim_type = torch.optim.Adam


@config('nes')
class NESConfig():
    n_rollout = 5
    n_step = 500
    population_size = 256
    seed = 123123


if __name__ == '__main__':
    nes = NES(config)

    @nes.optimize.add_hook()
    def after_optimize(self, *args, **kwargs):
        reward = self.eval_policy(self.policy)
        print(reward)

    nes.train()



from typing import Callable, List, Optional
from dataclasses import field
from itertools import tee

import gym
from pipcs import Config
import torch
from torch import nn
import numpy as np

from nes import NES, Policy, default_config, hook


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class Agent(Policy):
    def __init__(self, input_size, hidden_layers, output_size, output_func,
                 hidden_act, output_act=None, bias=True):
        super().__init__()
        layers: List[nn.Module] = []
        layer_sizes = [input_size] + hidden_layers + [output_size]
        for idx, (layer1, layer2) in enumerate(pairwise(layer_sizes)):
            layers.append(nn.Linear(layer1, layer2, bias=bias))
            if idx < len(layer_sizes) - 2:
                layers.append(hidden_act())
        if output_act is not None:
            layers.append(output_act())
        self.output_func = output_func
        self.seq = nn.Sequential(*layers)

    def rollout(self, env):
        done = False
        obs = env.reset()
        total_reward = 0
        while not done:
            obs_tensor = torch.from_numpy(obs).float()
            action = self.output_func(self.seq(obs_tensor))
            obs, reward, done, _ = env.step(action)
            total_reward += reward
        return total_reward



config = Config(default_config)


def make_env(id):
    return gym.make(id)


@config('environment')
class EnvironmentConfig():
    make_env = make_env
    id: str = 'LunarLander-v2'


@config('policy')
class PolicyConfig():
    policy = Agent
    env = NES.make_env(**config.environment.to_dict())
    input_size: int = env.observation_space.shape[0]
    if isinstance(env.action_space, gym.spaces.Discrete):
        output_size: int = env.action_space.n
        output_func: Callable[[int], int] = lambda x: x.argmax().item()
    else:
        output_size: int = env.action_space.shape[0]
        output_func: Callable[[int], int] = lambda x: x.detach().numpy()
    hidden_layers: List[int] = field(default_factory=lambda: [64, 32])
    hidden_act: nn.Module = nn.ReLU
    output_act: Optional[nn.Module] = nn.Tanh
    bias: bool = True


@config('optimizer')
class OptimizerConfig():
    lr = 0.02
    optim_type = torch.optim.Adam


@config('nes')
class NESConfig():
    n_rollout = 5
    n_step = 28
    l2_decay = 0.005
    population_size = 256
    sigma = 0.02
    seed = 123123


if __name__ == '__main__':
    @NES.__init__.add_hook()
    def init(self, *args, **kwargs):
        self.best_reward = -np.inf

    nes = NES(config)

    @nes.optimize.add_hook()
    def after_optimize(self, *args, **kwargs):
        reward = self.eval_policy(self.policy)
        if reward > self.best_reward:
            self.best_reward = reward
        print(f'Gen: {self.gen} Test Reward: {reward} Best Reward: {self.best_reward}')

    nes.train()

    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    if comm.Get_rank() == 0:

        nes.env = gym.wrappers.Monitor(nes.env, './videos', force = True)

        setattr(nes, 'env.step', hook(nes.env.step.__func__))
        @nes.env.step.add_hook(after=False)
        def render(self, *args, **kwargs):
            self.render()

        print(f'Reward: {nes.eval_policy(nes.policy)}')
